@@ -317,7 +317,7 @@ void gen_goto(code_lines_t &code, const State *from, const State *to
         code.push_back(o.str());
     }
 
-    gen_settags(code, dfa, tcid, opts);
+    gen_settags(code, dfa, tcid, opts, false /* delayed */);
 
     if (skip && opts->lookahead) {
         std::ostringstream o;
@@ -334,18 +334,24 @@ void gen_goto(code_lines_t &code, const State *from, const State *to
 void gen_on_eof_fail(code_lines_t &code, const opt_t *opts, const DFA &dfa
     , const State *from, const State *to, std::ostringstream &o)
 {
-    const State *fallback = from->rule == Rule::NONE
-        ? dfa.defstate : dfa.finstates[from->rule];
-    const tcid_t falltags = from->rule == Rule::NONE
-        ? from->fall_tags : from->rule_tags;
+    const bool final = from->rule != Rule::NONE;
+    const State *fallback = final ? dfa.finstates[from->rule] : dfa.defstate;
+    const tcid_t falltags = final ? from->rule_tags : from->fall_tags;
 
     if (from->action.type == Action::INITIAL) {
-        o << opts->indString << "goto " << opts->labelPrefix << "eof;";
+        o << opts->indString << "goto " << opts->labelPrefix;
+        if (final) {
+            o << fallback->label;
+        }
+        else {
+            o << "eof";
+        }
+        o << ";";
         flushln(code, o);
     }
     else if (fallback != to) {
         code_lines_t tagcode;
-        gen_settags(tagcode, dfa, falltags, opts);
+        gen_settags(tagcode, dfa, falltags, opts, false /* delayed */);
 
         if (tagcode.empty()) {
             o << opts->indString
@@ -382,7 +388,8 @@ void gen_on_eof(code_lines_t &code, const opt_t *opts, const DFA &dfa
         }
         flushln(code, o);
 
-        o << opts->indString << opts->fill << "();";
+        o << opts->indString << opts->fill
+            << (opts->fill_naked ? "" : opts->fill_arg_use ? "();" : ";");
         flushln(code, o);
 
         o << opts->indString << opts->labelPrefix << "eof" << fillidx << ":;";
@@ -392,8 +399,10 @@ void gen_on_eof(code_lines_t &code, const opt_t *opts, const DFA &dfa
     }
     else {
         if (opts->fill_use) {
-            o << opts->indString << "if (" << opts->fill << " () == 0) "
-                << "goto " << opts->labelPrefix << retry->label << "_;";
+            o << opts->indString << "if (" << opts->fill
+                << (opts->fill_naked ? ""
+                    : opts->fill_arg_use ? " () == 0" : " == 0")
+                << ") goto " << opts->labelPrefix << retry->label << "_;";
             flushln(code, o);
         }
         gen_on_eof_fail(code, opts, dfa, from, to, o);
@@ -403,7 +412,8 @@ void gen_on_eof(code_lines_t &code, const opt_t *opts, const DFA &dfa
     flushln(code, o);
 }
 
-void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid, const opt_t *opts)
+void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid,
+    const opt_t *opts, bool delayed)
 {
     const bool generic = opts->input_api == INPUT_CUSTOM;
     const std::string
@@ -413,6 +423,8 @@ void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid, const opt_t *o
 
     // single tag YYCTXMARKER, backwards compatibility
     if (cmd && dfa.oldstyle_ctxmarker) {
+        DASSERT(!opts->stadfa);
+
         const std::string s = generic
             ? opts->yybackupctx + " ();\n"
             : opts->yyctxmarker + " = " + opts->yycursor + ";\n";
@@ -444,7 +456,8 @@ void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid, const opt_t *o
             for (; *h != TAGVER_ZERO; ++h) {
                 const std::string s = *h == TAGVER_BOTTOM
                     ? opts->yymtagn + " (" + le + ");\n"
-                    : opts->yymtagp + " (" + le + ");\n";
+                    : (delayed ? opts->yymtagpd : opts->yymtagp)
+                        + " (" + le + ");\n";
                 code1.push_back(s);
             }
             code.insert(code.end(), code1.rbegin(), code1.rend());
@@ -455,7 +468,8 @@ void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid, const opt_t *o
                 v = vartag_expr(l, prefix, expression),
                 s = *h == TAGVER_BOTTOM
                     ? opts->yystagn + " (" + v + ");\n"
-                    : opts->yystagp + " (" + v + ");\n";
+                    : (delayed ? opts->yystagpd : opts->yystagp)
+                        + " (" + v + ");\n";
             code.push_back(s);
 
         // save command; no history; default API
@@ -470,7 +484,7 @@ void gen_settags(code_lines_t &code, const DFA &dfa, tcid_t tcid, const opt_t *o
                 code.push_back(s1);
             }
             if (!s2.empty()) {
-                s2 += opts->yycursor + ";\n";
+                s2 += opts->yycursor + (delayed ? " - 1" : "") + ";\n";
                 code.push_back(s2);
             }
         }
